# - Find GLib-2.0 library
# Find the native GLib headers and libraries.
#
#  GLIB_INCLUDE_DIRS - where to glib.h, glibconfig.h etc.
#  GLIB_LIBRARIES    - List of libraries when using GLib.
#  GLIB_FOUND        - True if Glib is found.

#=============================================================================
#Copyright 2000-2009 Kitware, Inc., Insight Software Consortium
#All rights reserved.
#
#Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#* Redistributions of source code must retain the above copyright notice, 
#this list of conditions and the following disclaimer.
#
#* Redistributions in binary form must reproduce the above copyright notice, 
#this list of conditions and the following disclaimer in the documentation 
#and/or other materials provided with the distribution.
#
#* Neither the names of Kitware, Inc., the Insight Software Consortium, nor 
#the names of their contributors may be used to endorse or promote products 
#derived from this software without specific prior written  permission.
#
#THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
#AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
#IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
#ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE 
#LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
#CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF 
#SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 
#INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
#CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
#ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
#POSSIBILITY OF SUCH DAMAGE.
#=============================================================================

# Look for the header file.
FIND_PATH(GLIB_INCLUDE_DIR NAMES glib.h 
  PATH_SUFFIXES glib-2.0
  DOC "The GLib include directory" )
MARK_AS_ADVANCED(GLIB_INCLUDE_DIR)

# glibconfig.h is sometimes stored in strange places. Use pkg-config to find it.
FIND_PACKAGE( PkgConfig )
IF( PKG_CONFIG_FOUND )
  PKG_CHECK_MODULES( GLIB2  glib-2.0 )
ENDIF()

FIND_PATH(GLIBCONFIG_INCLUDE_DIR NAMES glibconfig.h 
  PATH_SUFFIXES glib-2.0
  HINTS ${GLIB2_INCLUDE_DIRS}
  DOC "The glibconfig.h include directory" )
MARK_AS_ADVANCED(GLIBCONFIG_INCLUDE_DIR)

# Look for the library.
FIND_LIBRARY(GLIB_LIBRARY NAMES glib-2.0
  DOC "The GLib library" )
MARK_AS_ADVANCED(GLIB_LIBRARY)

# handle the QUIETLY and REQUIRED arguments and set GLIBFOUND_FOUND to TRUE if 
# all listed variables are TRUE
INCLUDE( ${CMAKE_ROOT}/Modules/FindPackageHandleStandardArgs.cmake )
FIND_PACKAGE_HANDLE_STANDARD_ARGS( GLib DEFAULT_MSG GLIB_LIBRARY GLIB_INCLUDE_DIR GLIBCONFIG_INCLUDE_DIR )

SET( GLIB_LIBRARIES ${GLIB_LIBRARY} )
SET( GLIB_INCLUDE_DIRS ${GLIB_INCLUDE_DIR} ${GLIBCONFIG_INCLUDE_DIR} )
